#!/usr/bin/env python
# -*- coding: utf-8 -*-
# nautilus-pastebin - Nautilus extension to paste a file to a pastebin service
# Written by:
#    Alessio Treglia <quadrispro@ubuntu.com>
# Copyright (C) 2009, Alessio Treglia
#
# Part of this code has been taken from pastebinit script
# pastebinit was written by: 
#    Stéphane Graber <stgraber@stgraber.org>
#    Daniel Bartlett <dan@f-box.org>
#    David Paleino <d.paleino@gmail.com>
# URL: http://www.stgraber.org/download/projects/pastebin/
# Copyright (C) 2006-2009, Stéphane Graber, Daniel Bartlett, David Paleino
# Distributed under the terms of the GNU GPL-2 license
#
# This package is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this package; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
#

import urllib
import gtk
import nautilus
import pynotify
import os, sys, re, sys, time
import mimetypes
import gettext

from ConfigParser import ConfigParser
from threading import Thread

gtk.gdk.threads_init() 

gettext.install("nautilus-pastebin")

# Globals
NAME_APP = "nautilus-pastebin"
GLOBALCONFIG_FILENAME = '/etc/nautilus-pastebin.conf'
USERCONFIG_FILENAME = os.path.expanduser('~/.config/nautilus-pastebin/nautilus-pastebin.conf')
PATH_APP = '/usr/share/nautilus-pastebin'
LOGGING = True

class ConfigFileException(Exception):
    pass

pynotify.init("Nautilus pastebin extension")

def log(message):
    """Log almost whatever happens."""
    if LOGGING:
        if sys.version.split()[0] < '3':
            print message
        else:
            print (message)

# Custom urlopener to handle 401's
class pasteURLopener(urllib.FancyURLopener):
    """Do the HTTP requests."""
    def http_error_401(self, url, fp, errcode, errmsg, headers, data=None):
            return None

class GlobalConfigurationValidator(object):
    """Configuration wrapper """
    NEEDED_OPTS = ['pastebin']
    
    def validate(self):
        log ("Validating configuration files...")
        conf = {}
        # Looking for a valid [generals] section
        opts = self.__config.options('generals')
        for k in opts:
            conf[k] = self.__config.get('generals', k)
            # Each needed field must be fill'd.
            if k in GlobalConfigurationValidator.NEEDED_OPTS and conf[k] == '':
                raise ConfigFileException, "Missing required value for %s field!" % k
        # Everything seems well-configured
        for k,v in conf.items():
            setattr(self,k,v)
        return True

    def getPastebinConfiguration(self):
        """Retrieve and returns the configuration related to the pastebin service."""
        pastebinConfiguration = None
        return PastebinConfiguration(dict(self.__config.items(self.pastebin)))
    
    def __init__(self):
        log ("Initializing global configuration validator...")
        self.__config = ConfigParser()
        self.__config.read([GLOBALCONFIG_FILENAME, USERCONFIG_FILENAME])

class PastebinConfiguration(object):
    PARAMETERS = ["website",
        "content",
        "user",
        "jabberid",
        "version",
        "format",
        "parentid",
        "permatag",
        "title",
        "username",
        "password",
        "default_mime"]

    def __setDefaultParameters(self):
        for k in PastebinConfiguration.PARAMETERS:
            self.__dict__.setdefault(k,'')
        for k in PastebinConfiguration.PARAMETERS:
            if k in self._optionsDict.keys():
                setattr(self, k, self._optionsDict[k])

    def __getMappingByName(self, optname):
        map = {}
        if optname in self._optionsDict:
            exec(self._optionsDict[optname])
        return map

    def __init__(self, optionsDict):
        log ("Initializing pastebin configuration...")
        self._optionsDict = optionsDict
        self.__setDefaultParameters()
        self.aliases = self.__getMappingByName('aliases')
        self.extras = self.__getMappingByName('extras')
        self.mimetypes = self.__getMappingByName('mimetypes')
        self.mimetypes.setdefault('text/plain', self._optionsDict['default_mime'])

    def setFormatByMimeType(self, mimetype):
        if mimetype in self.mimetypes.keys():
            self.format = self.mimetypes[mimetype]
        elif mimetype.split('/')[0] == 'text':
            self.format = self.mimetypes['text/plain']
        else:
            return False
        return True
    
    def __str__(self):
        s = 'Configuration:\n'
        s += "mimetypes = %s\n" % self.mimetypes.__str__()
        s += "aliases = %s" % self.aliases.__str__()
        s += "extras = %s" % self.extras.__str__()
        return s

    def setContent(self, pastefile):
        log ("Trying to read %s" % pastefile)
        fp = open(pastefile)
        content = fp.read()
        fp.close()
        if not content:
            log ("Don't send empty content.")
            return False
        # OK, let's go...
        self.content = content
        self.title = os.path.split(pastefile)[-1]
        return True

class PastebinValidatorFactory(object):
    """Singleton factory which returns a validator object dinamically."""
    __instance = None
    class __impl(object):
        """Implementation of the class' methods"""
        def getValidator(self, pastebinConfiguration):
            website = pastebinConfiguration.website
            log ("Looking for a validator: %s" %website)
            if (re.search("http://((([a-zA-Z0-9\-_\.]*)(pastebin\.com)))", website) and not website == "http://www.pastebin.com") or website == "http://pastebin.mozilla.org":
                return PastebinComValidator(pastebinConfiguration)
            elif re.search("http://((([a-zA-Z0-9\-_\.]*)(pastebin\.ca)))", website):
                return PastebinCaValidator(pastebinConfiguration)
            elif website == "http://stikked.com":
                return StikkedComValidator(pastebinConfiguration)
            elif website == "http://slexy.org":
                return SlexyOrgValidator(pastebinConfiguration)
            elif website == "http://fpaste.org":
                return FpasteValidator(pastebinConfiguration)
            elif website == "http://paste2.org":
                return Paste2OrgValidator(pastebinConfiguration)
            elif re.search("http://((([a-zA-Z0-9\-_\.]*)(pastey\.net)))", website):    
                return PasteyNetValidator(pastebinConfiguration)
            elif website == "http://yourpaste.net":
                return YourpasteNetValidator(pastebinConfiguration)
            elif website == "http://gist.github.com":
                return GistGithubComValidator(pastebinConfiguration)
            elif website == "http://paste.ubuntu.com":
                return PasteUbuntuComValidator(pastebinConfiguration)
            elif website == "http://paste.debian.net":
                return PasteDebianNetValidator(pastebinConfiguration)
            else:
                return None # Service has not been recognized
    def __init__(self):
        if PastebinValidatorFactory.__instance is None:
            PastebinValidatorFactory.__instance = PastebinValidatorFactory.__impl()
        self.__dict__['_PastebinValidatorFactory__instance'] = PastebinValidatorFactory.__instance
    def __getattr__(self, attr):
        return getattr(self.__instance, attr)
    def __setattr__(self, attr, value):
        return setattr(self.__instance, attr, value)

class GenericPastebinValidator(object):
    NOT_EMPTY_FIELDS = ['website', 'content']

    def checkAliases(self):
        return all(
            map(
                lambda x: x in self.conf.__dict__, self.conf.aliases.keys())
                )

    def checkExtras(self): return True

    def __init__(self, pastebinConfiguration):
        """Class constructor"""
        log ("New generic validator constructor.")
        cls = self.__class__
        if cls.__name__ != 'GenericPastebinValidator':
            try:
                toAdd = getattr(cls, 'notEmptyFields')
                GenericPastebinValidator.NOT_EMPTY_FIELDS.extend(toAdd)
            except:
                return
        self.conf = pastebinConfiguration

    def validate(self):
        """Do the validation of the configuration file."""
        if not self.checkNotEmpty() or not self.checkAliases() or not self.checkExtras():
            return False
        if self.conf.user == '':
            self.conf.user = os.environ.get('USER')
        return True

    def checkNotEmpty(self):
        """Verify whether all the opts' keys are non-empty."""
        try:
            for i in GenericPastebinValidator.NOT_EMPTY_FIELDS:
                if getattr(self.conf, i) == '':
                    raise Exception
        except:
            return False
        return True

    def getParameters(self):
        params = {}
        # Set the aliases
        for k in self.conf.aliases.keys():
            for v in self.conf.aliases[k]:
                params[v] = getattr(self.conf, k)
        for k in self.conf.extras.keys():
            params[k] = self.conf.extras[k]
        return params

class PasteUbuntuComValidator(GenericPastebinValidator):
    notEmptyFields = ['format']

class PasteDebianNetValidator(GenericPastebinValidator):
    notEmptyFields = ['format']

class PastebinComValidator(GenericPastebinValidator):
    notEmptyFields = ['format']

class GistGithubComValidator(GenericPastebinValidator):
    notEmptyFields = ['format', 'title']

class YourpasteNetValidator(GenericPastebinValidator):
    notEmptyFields = ['format', 'title']

class PasteyNetValidator(GenericPastebinValidator):
    notEmptyFields = ['format', 'title']
    def parentFixup(self, website, parentid):
        if parentid == "":
            return ""
        url_opener = pasteURLopener()
        page = url_opener.open(website + '/' + parentid, None)
        matches = re.split('<input.*?name="parent".*?value="(.*?)"', page.read())
        if len(matches) <= 1 or re.match(parentid, matches[1]) == None:
            # The obfuscated version didn't begin with the partial version,
            # or unable to find the obfuscated version for some reason!
            # Create a paste with no parent (should we throw, instead?)
            return ""
        return matches[1]
    def getParameters(self):
        params = GenericPastebinValidator.getParameters(self)
        params['parent'] = self.parentFixup(self.conf.website, self.conf.parentid)
        return params

class PastebinCaValidator(GenericPastebinValidator):
    notEmptyFields = ['format']

class StikkedComValidator(GenericPastebinValidator):
    notEmptyFields = ['format', 'title']
    def getParameters(self):
        params = GenericPastebinValidator.getParameters(self)
        params['website'] = ''
        params['email'] = ''
        return params

class SlexyOrgValidator(GenericPastebinValidator):
    notEmptyFields = ['format', 'title']
    def getParameters(self):
        params = GenericPastebinValidator.getParameters(self)
        params['comment'] = ''
        return params

class FpasteValidator(GenericPastebinValidator):
    notEmptyFields = ['format']

class Paste2OrgValidator(GenericPastebinValidator):
    notEmptyFields = ['format', 'title']

class PastebinThread(Thread):
    def __init__(self, pasteconf):
        self.pasteconf = pasteconf
        Thread.__init__(self)

    def run(self):
        log ("PastebinThread started!")
        validator = PastebinValidatorFactory().getValidator(self.pasteconf)
        if not validator.validate():
            return
        website = self.pasteconf.website
        params = validator.getParameters()

        if not re.search(".*/", website):
            log ("Regex found")
            website += "/"

        reLink=None
        tmp_page=""
        if params.__contains__("page"):
            website+=params['page']
            tmp_page=params['page']
            params.__delitem__("page")
        if params.__contains__("regexp"):
            reLink=params['regexp']
            params.__delitem__("regexp")
        params = urllib.urlencode(params) #Convert to a format usable with the HTML POST

        try:
            url_opener = pasteURLopener()
            page = url_opener.open(website, params) #Send the informations and be redirected to the final page
            summary = 'File pasted to: '
            URLmarkup = '<a href="%(pasteurl)s">%(pasteurl)s</a>'
            if reLink: #Check if we have to apply a regexp
                website = website.replace(tmp_page, "")
                pasteurl = website + "/" + re.split(reLink, page.read())[1]
            else:
                pasteurl = page.url
            message = URLmarkup % { 'pasteurl': pasteurl }
            # Paste to the clipboard, too
            log ("Pasting to the clipboard...")
            cb = gtk.clipboard_get('CLIPBOARD')
            cb.clear()
            cb.set_text(pasteurl)
            #cb.store()
            log ("Done.")
        except Exception, value:
            summary = _("Unable to read or parse the result page")
            message = _("It could be a server timeout or a change server side, try later.")
            log ("Exception: %s" % value)
        finally:
            n = pynotify.Notification(summary, message)
            n.show()

class PastebinitExtension(nautilus.MenuProvider):
    def __init__(self):
        log ("Intializing nautilus-pastebin extension...")
    
    def get_file_items(self, window, files):
        if len(files)!=1:
            return
        filename = files[0]
        
        validator = GlobalConfigurationValidator()
        validator.validate()
        pasteconf = validator.getPastebinConfiguration()
                
        if not pasteconf.setFormatByMimeType(filename.get_mime_type()):
            log ("Unable to found a mapping for the MIME type.")
            return
        self.pasteconf = pasteconf

        items = []
        #Called when the user selects a file in Nautilus.
        item = nautilus.MenuItem("NautilusPython::pastebin_item",
                                 _("Pastebin"),
                                 _("Pastebin"))
        item.set_property('icon', "nautilus-pastebin")
        item.connect("activate", self.menu_activate_cb, files)
        items.append(item)
        return items

    def menu_activate_cb(self, menu, files):
        if len(files) != 1:
            return
        filename = files[0]
        if filename.get_uri_scheme() != 'file':
            return

        if filename.is_directory():
            return

        pastefile = urllib.unquote(filename.get_uri()[7:])
        if not self.pasteconf.setContent(pastefile):
            return
        
        thread = PastebinThread(self.pasteconf)
        thread.start()
        while (thread.isAlive()):
            time.sleep(0.09)
            while gtk.events_pending():
                gtk.main_iteration()

